package java.util.concurrent.atomic;

import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.util.function.DoubleBinaryOperator;
import java.io.Serializable;

public class DoubleAccumulator extends Striped64 implements Serializable
{
    private static final long serialVersionUID = 7249069246863182397L;
    private final DoubleBinaryOperator function;
    private final long identity;
    
    public DoubleAccumulator(DoubleBinaryOperator function, double n) {
        this.function = function;
        long doubleToRawLongBits = Double.doubleToRawLongBits(n);
        this.identity = doubleToRawLongBits;
        this.base = doubleToRawLongBits;
    }
    
    public void accumulate(double n) {
        Striped64.Cell[] cells;
        long base;
        long doubleToRawLongBits;
        if ((cells = this.cells) != null || ((doubleToRawLongBits = Double.doubleToRawLongBits(this.function.applyAsDouble(Double.longBitsToDouble(base = this.base), n))) != base && !this.casBase(base, doubleToRawLongBits))) {
            boolean n2 = true;
            int n3;
            Striped64.Cell cell;
            long value;
            long doubleToRawLongBits2;
            if (cells == null || (n3 = cells.length - 1) < 0 || (cell = cells[getProbe() & n3]) == null || (n2 = ((doubleToRawLongBits2 = Double.doubleToRawLongBits(this.function.applyAsDouble(Double.longBitsToDouble(value = cell.value), n))) == value || cell.cas(value, doubleToRawLongBits2))) == 0) {
                this.doubleAccumulate(n, this.function, (boolean)n2);
            }
        }
    }
    
    public double get() {
        Striped64.Cell[] cells = this.cells;
        double n = Double.longBitsToDouble(this.base);
        if (cells != null) {
            for (int i = 0; i < cells.length; ++i) {
                Striped64.Cell cell;
                if ((cell = cells[i]) != null) {
                    n = this.function.applyAsDouble(n, Double.longBitsToDouble(cell.value));
                }
            }
        }
        return n;
    }
    
    public void reset() {
        Striped64.Cell[] cells = this.cells;
        this.base = this.identity;
        if (cells != null) {
            for (int i = 0; i < cells.length; ++i) {
                Striped64.Cell cell;
                if ((cell = cells[i]) != null) {
                    cell.value = this.identity;
                }
            }
        }
    }
    
    public double getThenReset() {
        Striped64.Cell[] cells = this.cells;
        double n = Double.longBitsToDouble(this.base);
        this.base = this.identity;
        if (cells != null) {
            for (int i = 0; i < cells.length; ++i) {
                Striped64.Cell cell;
                if ((cell = cells[i]) != null) {
                    double longBitsToDouble = Double.longBitsToDouble(cell.value);
                    cell.value = this.identity;
                    n = this.function.applyAsDouble(n, longBitsToDouble);
                }
            }
        }
        return n;
    }
    
    @Override
    public String toString() {
        return Double.toString(this.get());
    }
    
    @Override
    public double doubleValue() {
        return this.get();
    }
    
    @Override
    public long longValue() {
        return (long)this.get();
    }
    
    @Override
    public int intValue() {
        return (int)this.get();
    }
    
    @Override
    public float floatValue() {
        return (float)this.get();
    }
    
    private Object writeReplace() {
        return new DoubleAccumulator.SerializationProxy(this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }
    
    private static class SerializationProxy implements Serializable
    {
        private static final long serialVersionUID = 7249069246863182397L;
        private final double value;
        private final DoubleBinaryOperator function;
        private final long identity;
        
        SerializationProxy(DoubleAccumulator doubleAccumulator) {
            this.function = doubleAccumulator.function;
            this.identity = doubleAccumulator.identity;
            this.value = doubleAccumulator.get();
        }
        
        private Object readResolve() {
            DoubleAccumulator doubleAccumulator = new DoubleAccumulator(this.function, Double.longBitsToDouble(this.identity));
            doubleAccumulator.base = Double.doubleToRawLongBits(this.value);
            return doubleAccumulator;
        }
    }
}
